from subprocess import Popen, PIPE

from django.utils.datastructures import SortedDict
from django.utils.functional import memoize

from .utils import get_class

_filters = SortedDict()
_processes = SortedDict()


class BaseFilter(object):
    method = None
    options = {}

    def __init__(self, *args, **kwargs):
        for option, default in self.options.items():
            setattr(self, option, kwargs[option] if option in kwargs else default)
        self.options.update(**kwargs)

    def __call__(self, asset):
        if not self.method:
            raise Exception("'%s' filter must specify a method to execute" % self.__class__.__name__)

        method = getattr(self, self.method)
        if method:
            method(asset)


class CommandMixin(object):
    command = None
    params = []

    def get_args(self):
        return [self.options.get('command', self.command)] + self.options.get('params', self.params)

    def run(self, input):
        args = self.get_args()
        process = Popen(args, stdin=PIPE, stdout=PIPE, stderr=PIPE)
        output, errors = process.communicate(input=input.encode('utf-8'))
        if process.returncode != 0:
            raise Exception("'%s'\n%s" % (' '.join(args), errors))
        return output.decode('utf-8')


def get_filter(filter_conf):
    if isinstance(filter_conf, (tuple, list)):
        filter_conf = list(filter_conf)
    else:
        filter_conf = [filter_conf]
    import_path = filter_conf.pop(0)
    args, kwargs = get_arguments(filter_conf)
    key = ','.join(args) + ','.join(['%s:%s' % (k, v) for k, v in kwargs.items()])
    return get_cached_filter(import_path, key, args, kwargs)


def _get_filter(import_path, key, args, kwargs):
    return get_class(import_path, BaseFilter)(*args, **kwargs)
get_cached_filter = memoize(_get_filter, _filters, 2)


def _get_process(key, args):
    print _processes
    return Popen(args, stdin=PIPE, stdout=PIPE, stderr=PIPE)
get_process = memoize(_get_process, _processes, 1)


def get_arguments(conf):
    args, kwargs = [], {}
    for arg in conf:
        if isinstance(arg, dict):
            kwargs = arg
        elif isinstance(arg, (tuple, list)):
            args = arg
        else:
            args.append(arg)
    return args, kwargs



import os

from django.core.files.base import ContentFile
from django.contrib.staticfiles.storage import StaticFilesStorage

from .finder import default_finder as finder
from . import settings

class StaticAssetsStorage(StaticFilesStorage):

    def post_process(self, paths, dry_run=False, **options):
        if dry_run:
            return []
        try:
            manifests = settings.STATICASSETS_MANIFESTS
        except AttributeError:
            print "you have to define the STATICASSETS_MANIFESTS in the django settings"
            return []
        for manifest in manifests:
            asset = finder.find(manifest, bundle=True)
            try:
                os.remove(self.path(asset.name))
            except OSError:
                pass
            self.save(asset.name, ContentFile(asset.content))
        return []



import mimetypes

from django.http import HttpResponse, HttpResponseNotModified, Http404
from django.contrib.staticfiles.views import serve as staticfiles_serve
from django.views.static import was_modified_since, http_date

from staticassets import finder, settings


def serve(request, path, **kwargs):
    mimetype, encoding = mimetypes.guess_type(path)
    if not mimetype in settings.MIMETYPES.values():
        return staticfiles_serve(request, path, **kwargs)

    bundle = request.GET.get('bundle') in ('1', 't', 'true') or not settings.DEBUG
    asset = finder.find(path, bundle=bundle)
    if not asset:
        raise Http404("Static asset not found")

    # Respect the If-Modified-Since header.
    modified_since = request.META.get('HTTP_IF_MODIFIED_SINCE')
    if not was_modified_since(modified_since, asset.mtime, asset.size):
        return HttpResponseNotModified(content_type=asset.attributes.content_type)

    response = HttpResponse(asset.content, content_type=asset.attributes.content_type)
    response['Last-Modified'] = http_date(asset.mtime)
    response['Content-Length'] = asset.size

    return response



from django.contrib.staticfiles.handlers import StaticFilesHandler

from . import views


class AssetsHandler(StaticFilesHandler):
    def serve(self, request):
        return views.serve(request, self.file_path(request.path), insecure=True)



import os

from django.conf import settings


DEBUG = getattr(settings, 'STATICASSETS_DEBUG', settings.DEBUG)

MIMETYPES = {
    '.css': 'text/css',
    '.js': 'application/javascript'
}
MIMETYPES.update(**getattr(settings, 'MIMETYPES', {}))

DIRS = getattr(settings, 'STATICASSETS_DIRS', getattr(settings, 'STATICFILES_DIRS'))

CACHE_DIR = os.path.join(getattr(settings, 'STATIC_ROOT', '/tmp'), 'staticassets-cache')

FINDER = getattr(settings, 'STATICASSETS_FINDER', 'staticassets.finder.AssetFinder')

PREPROCESSORS = getattr(settings, 'STATICASSETS_PREPROCESSORS', (
    ('application/javascript', 'staticassets.processors.DirectiveProcessor'),
    ('text/css', 'staticassets.processors.DirectiveProcessor')
))

POSTPROCESSORS = getattr(settings, 'STATICASSETS_POSTPROCESSORS', tuple())

COMPILERS = {
    '.sass': 'staticassets.compilers.SassCompiler',
    '.scss': 'staticassets.compilers.SassCompiler',
    '.styl': 'staticassets.compilers.StylusCompiler',
    '.less': 'staticassets.compilers.LessCompiler',
    '.jst': 'staticassets.compilers.JstCompiler',
    '.ejs': 'staticassets.compilers.EjsCompiler',
    '.coffee': 'staticassets.compilers.CoffeeScriptCompiler'
}
COMPILERS.update(**getattr(settings, 'STATICASSETS_COMPILERS', {}))

AVAILABLE_EXTENSIONS = MIMETYPES.keys() + COMPILERS.keys()

STATICASSETS_MANIFESTS = getattr(settings, 'STATICASSETS_MANIFESTS', ())



import os
import codecs
from setuptools import setup

README = codecs.open(os.path.join(os.path.dirname(__file__), 'README.rst')).read()

# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

setup(
    name='django-staticassets',
    version='0.1',
    packages=['staticassets'],
    include_package_data=True,
    license='MIT License',
    description='Compiles and bundles static assets. Uses a directive processor similiar to Ruby\'s Sprockets',
    long_description=README,
    url='https://github.com/davidelias/django-staticassets',
    author='David Elias',
    author_email='david@david-elias.net',
    classifiers=[
        'Environment :: Web Environment',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
    ],
)



import os
import codecs
import hashlib
import json

from django.core.exceptions import ImproperlyConfigured
from django.utils.importlib import import_module


def read_file(path):
    with codecs.open(path, 'r', encoding='utf-8') as f:
        return f.read()


def get_digest(source):
    return hashlib.md5(source.encode('utf-8')).hexdigest()


def get_path_digest(path):
    if os.path.isfile(path):
        return get_digest(read_file(path))
    elif os.path.isdir(path):
        return get_digest(','.join(os.listdir(path)))


def expand_component_json(self, path):
    component = json.loads(read_file(path))
    if not 'main' in component:
        yield
    elif isinstance(component['main'], list):
        for name in component['main']:
            yield name
    else:
        yield component['main']


def get_class(import_path, class_type):
    """
    """
    module, attr = import_path.rsplit('.', 1)
    try:
        mod = import_module(module)
    except ImportError as e:
        raise ImproperlyConfigured('Error importing module %s: "%s"' % (module, e))
    try:
        cls = getattr(mod, attr)
    except AttributeError:
        raise ImproperlyConfigured('Module "%s" does not define a "%s" class.' % (module, attr))
    if not issubclass(cls, class_type):
        raise ImproperlyConfigured('Filter "%s" is not a subclass of "%s"' % (cls, class_type))
    return cls



import os
import re

from django.core.cache import get_cache
from django.core.files.storage import FileSystemStorage
from django.contrib.staticfiles import finders
from django.utils.datastructures import SortedDict
from django.utils.functional import SimpleLazyObject, memoize

from staticassets import Asset, AssetAttributes
from .exceptions import AssetNotFound
from .utils import expand_component_json, get_class
from . import settings


cache = SimpleLazyObject(lambda: get_cache('django.core.cache.backends.filebased.FileBasedCache', **{
    'LOCATION': os.path.join(settings.CACHE_DIR),
    'TIMEOUT': 60 * 60 * 24 * 30
}))


class BaseAssetFinder(object):
    """
    A base asset finder to be used for custom asset finder classes.
    """
    def find(self, path, bundle=False, **options):
        raise NotImplementedError()


class AssetFinder(BaseAssetFinder):
    """
    Finder to locate any asset, it will use app `django.contrib.staticfiles`
    finders and allow search with or without extensions
    """

    def __init__(self):
        # cache in memory
        self.assets = {}
        self.search_regex = {}

    def find(self, path, **options):
        options.setdefault('bundle', False)
        name, storage = self.resolve(path, **options)
        key = self.get_cache_key(storage.path(name), options)
        asset = self.assets.get(key, cache.get(key))
        if not asset or asset.expired:
            asset = self.assets[key] = Asset.create(name, storage, **options)
            cache.set(key, asset)
        return asset

    def resolve(self, path, **options):
        # first try to match the exact filename
        absolute_path = finders.find(path)
        if absolute_path and os.path.isfile(absolute_path):
            return path, FileSystemStorage(location=absolute_path[:-len(path)])

        attrs = AssetAttributes(path)
        for search_path in attrs.search_paths:
            regex = self.get_search_regex(search_path, options.get('content_type'))
            for name, storage in self.list():
                if search_path.endswith('component.json') and search_path == name:
                    for name in expand_component_json(storage.path(name)):
                        return name, storage
                elif regex.search(name):
                    return name, storage

        raise AssetNotFound('File "%s" not found. Tried "%s"' % (
            attrs.path, '", "'.join(attrs.search_paths)))

    def list(self):
        for finder in finders.get_finders():
            for result in finder.list(None):
                yield result

    def get_search_regex(self, path, content_type=None):
        if not (path, content_type) in self.search_regex:
            attrs = AssetAttributes(path, content_type)
            # remove all extensions except non compiler and mimetype extensions
            path = attrs.path_without_extensions + attrs.suffix
            pattern = '|'.join([r'\%s' % ext for ext in attrs.available_extensions])
            self.search_regex[(path, content_type)] = re.compile(r'^%s(%s)*$' % (path, pattern))
        return self.search_regex[(path, content_type)]

    def get_cache_key(self, path, options):
        return '%s:bundle:%s' % (path, 1 if options.get('bundle') else 0)

StaticFilesFinder = AssetFinder


def find(*args, **kwargs):
    return default_finder.find(*args, **kwargs)


default_finder = SimpleLazyObject(lambda: get_finder(settings.FINDER))


_finders = SortedDict()


def _get_finder(import_path):
    return get_class(import_path, BaseAssetFinder)()
get_finder = memoize(_get_finder, _finders, 1)
